﻿
Imports System.Globalization
Imports System.Math
Imports System.Collections.Generic

' ==================================================================================================
'   CLASS SLAVE
' ==================================================================================================
Class Slave

    Enum SlaveTypes As Int32
        Custom = 0
        CapSensor = 1
        Servo = 2
        Generic = 3
        InOut = 4
        MasterPins = 5
        'RmsLogger = 6
        'Sniffer = 7
        MasterPinsV2 = 8
        MasterPinsV4 = 9
        Unknown = 255
    End Enum

    Friend MasterId As Int32
    Friend mSlaveId As Int32
    Friend SlaveType As SlaveTypes
    Friend Pins As List(Of Pin)
    Friend LoadCurrent_mA As Int32

    Friend Property SlaveId() As Int32
        Get
            Return mSlaveId
        End Get
        Set(ByVal value As Int32)
            ' test if mSlaveId ok
            mSlaveId = value
        End Set
    End Property

    Friend Function SlaveTypeString() As String
        Select Case SlaveType
            Case SlaveTypes.Custom : Return "Custom"
            Case SlaveTypes.CapSensor : Return "CapSensor"
            Case SlaveTypes.Servo : Return "Servo"
            Case SlaveTypes.Generic : Return "Generic"
            Case SlaveTypes.InOut : Return "InOut"
            Case SlaveTypes.MasterPins : Return "MasterPins"
            Case SlaveTypes.MasterPinsV2 : Return "MasterPins"
            Case SlaveTypes.MasterPinsV4 : Return "MasterPins"
            Case SlaveTypes.Unknown : Return "Unknown"
            Case Else : Return "Unknown"
        End Select
    End Function

    Friend Sub New(ByVal _slavetype As SlaveTypes, ByVal _masterid As Int32, ByVal _slaveid As Int32)
        SlaveType = _slavetype
        MasterId = _masterid
        SlaveId = _slaveid
        LoadCurrent_mA = 0
    End Sub

    Friend Function GetHostToMasterBytesCount() As Int32
        GetHostToMasterBytesCount = 0
        For Each p As Pin In Pins
            If p.Direction = Pin.Directions.HostToMaster Then
                GetHostToMasterBytesCount += p.UsbBytesCount
            End If
        Next
    End Function

    Friend Function GetMasterToHostBytesCount() As Int32
        GetMasterToHostBytesCount = 0
        For Each p As Pin In Pins
            If p.Direction = Pin.Directions.MasterToHost Then
                GetMasterToHostBytesCount += p.UsbBytesCount
            End If
        Next
    End Function

    Friend Function CalibrationNeeded() As Boolean
        For Each p As Pin In Pins
            Select Case p.GetPinType()
                Case Pin.PinTypes.CAP_SENSOR
                    If p.Area_cmq > 0 Then
                        Return True
                    End If
                Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                    If p.MinVariation > 0 Then
                        Return True
                    End If
            End Select
        Next
        Return False
    End Function

    Friend Overridable Sub Calibrate()
        For Each p As Pin In Pins
            p.Calibrate()
        Next
    End Sub

    Friend Overridable Sub CalibrateZero()
        For Each p As Pin In Pins
            p.CalibrateZero()
        Next
    End Sub

    Friend Overridable Function Get_CapSensor_Or_Cap_Delta() As UInt32
        Dim maxdelta As UInt32 = 0
        For Each p As Pin In Pins
            Select Case p.GetPinType()
                Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                    Dim n As UInt32 = AbsUint32Difference(p.Value_Uinteger_Max, p.Value_Uinteger_Min)
                    If n > maxdelta Then maxdelta = n
                    p.ResetPosMinMax()
            End Select
        Next
        Return maxdelta
    End Function

    Friend Overridable Sub ConfigurePin(ByVal npin As Int32, ByVal pinType As Pin.PinTypes)
        Select Case SlaveType
            Case SlaveTypes.Servo, SlaveTypes.Generic, SlaveTypes.InOut

                Select Case pinType

                    Case Pin.PinTypes.UNUSED, Pin.PinTypes.DIG_OUT, _
                                                Pin.PinTypes.PWM_8, _
                                                Pin.PinTypes.PWM_16, _
                                                Pin.PinTypes.SERVO_8, _
                                                Pin.PinTypes.SERVO_16, _
                                                Pin.PinTypes.DIG_IN, _
                                                Pin.PinTypes.DIG_IN_PU, _
                                                Pin.PinTypes.COUNTER, _
                                                Pin.PinTypes.COUNTER_PU
                        Pins(npin).SetPinType(pinType)

                    Case Pin.PinTypes.ADC_8, Pin.PinTypes.ADC_16, _
                         Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16, _
                         Pin.PinTypes.RES_8, Pin.PinTypes.RES_16
                        If npin < 8 Then Pins(npin).SetPinType(pinType)

                    Case Pin.PinTypes.FAST_COUNTER, Pin.PinTypes.FAST_COUNTER_PU
                        If npin = 7 Then Pins(npin).SetPinType(pinType)

                    Case Pin.PinTypes.USOUND_SENSOR, Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                        If npin = 8 Then Pins(npin).SetPinType(pinType)
                End Select
        End Select
    End Sub

    Friend Overridable Sub FillTypeCombo(ByVal cmb As ComboBox, ByRef p As Pin)
        Select Case SlaveType

            Case SlaveTypes.Servo, SlaveTypes.Generic, SlaveTypes.InOut
                cmb.Items.Clear()
                '
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED))
                '
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16))
                '
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU))
                '
                If p.PinId < 8 Then
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16))
                End If
                '
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU))

                If p.PinId = 7 Then
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU))
                End If

                If p.PinId = 8 Then
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU))
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR))
                End If

            Case Else
                cmb.Items.Clear()
                cmb.Items.Add(Pin.PinTypeToString(p.GetPinType))

        End Select
    End Sub

    Friend Function ValidLinkedToPrevious(ByVal pinSelected As Pin) As Boolean
        If Not pinSelected.GetPinType() = Pin.PinTypes.STEPPER Then Return False
        If pinSelected.PinId < 2 Then Return False
        If Pins(pinSelected.PinId - 2).GetPinType() <> Pin.PinTypes.STEPPER Then Return False
        Return False 'True  ' <<<<< TODO  
    End Function

End Class


' ==================================================================================================
' SUBCLASS - Slave_CapSensor
' ==================================================================================================
Class Slave_CapSensor : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.CapSensor, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 12
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
    End Sub

    Friend Overrides Function Get_CapSensor_Or_Cap_Delta() As UInt32
        With Pins(0)
            Dim n As UInt32 = AbsUint32Difference(.Value_Uinteger_Max, .Value_Uinteger_Min)
            If n < 0 Then n = 0
            If n > 400000 Then n = 0
            Get_CapSensor_Or_Cap_Delta = n
            .ResetPosMinMax()
        End With
    End Function

    Friend Overrides Sub ConfigurePin(ByVal npin As Int32, ByVal pinType As Pin.PinTypes)
        Select Case pinType
            Case Pin.PinTypes.UNUSED, Pin.PinTypes.CAP_SENSOR
                Pins(npin).SetPinType(pinType)
        End Select
    End Sub

    Friend Overrides Sub FillTypeCombo(ByVal cmb As ComboBox, ByRef p As Pin)
        cmb.Items.Clear()
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_SENSOR))
    End Sub

End Class


' ==================================================================================================
'   SUBCLASS - Slave_Servo
' ==================================================================================================
Class Slave_Servo : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.Servo, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 80
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9))
    End Sub

End Class


' ==================================================================================================
'   SUBCLASS - Slave_Generic
' ==================================================================================================
Class Slave_Generic : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.Generic, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 80
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11))
    End Sub

End Class


' ==================================================================================================
'   SUBCLASS - Slave_InOut
' ==================================================================================================
Class Slave_InOut : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.InOut, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 80
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11))
    End Sub

End Class


' ==================================================================================================
'   SUBCLASS - Slave_MasterPins
' ==================================================================================================
Class Slave_MasterPins : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.MasterPins, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 10
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5))
    End Sub

    Friend Overrides Sub ConfigurePin(ByVal npin As Int32, ByVal pinType As Pin.PinTypes)
        Select Case pinType

            Case Pin.PinTypes.UNUSED, Pin.PinTypes.DIG_OUT, Pin.PinTypes.PWM_8, _
                                                            Pin.PinTypes.PWM_16, _
                                                            Pin.PinTypes.SERVO_8, _
                                                            Pin.PinTypes.SERVO_16, _
                                                            Pin.PinTypes.DIG_IN, _
                                                            Pin.PinTypes.DIG_IN_PU, _
                                                            Pin.PinTypes.ADC_8, _
                                                            Pin.PinTypes.ADC_16, _
                                                            Pin.PinTypes.CAP_8, _
                                                            Pin.PinTypes.CAP_16, _
                                                            Pin.PinTypes.RES_8, _
                                                            Pin.PinTypes.RES_16, _
                                                            Pin.PinTypes.COUNTER, _
                                                            Pin.PinTypes.COUNTER_PU
                Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.FAST_COUNTER, Pin.PinTypes.FAST_COUNTER_PU
                If Not AlreadyUsed_FastCounter(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.USOUND_SENSOR, Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                If Not AlreadyUsed_PeriodCounter(Pins(npin)) Then Pins(npin).SetPinType(pinType)
        End Select
    End Sub

    Friend Overrides Sub FillTypeCombo(ByVal cmb As ComboBox, ByRef p As Pin)
        cmb.Items.Clear()
        '
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED))
        '
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16))
        '
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER))
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU))

        If Not AlreadyUsed_FastCounter(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU))
        End If

        If Not AlreadyUsed_PeriodCounter(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR))
        End If
    End Sub

    Private Function AlreadyUsed_FastCounter(ByVal pinSelected As Pin) As Boolean
        For Each p As Pin In Pins
            If p Is pinSelected Then Continue For
            If p.GetPinType = Pin.PinTypes.FAST_COUNTER OrElse p.GetPinType = Pin.PinTypes.FAST_COUNTER_PU Then
                Return True
            End If
        Next
        Return False
    End Function

    Private Function AlreadyUsed_PeriodCounter(ByVal pinSelected As Pin) As Boolean
        For Each p As Pin In Pins
            If p Is pinSelected Then Continue For
            If p.GetPinType = Pin.PinTypes.PERIOD OrElse p.GetPinType = Pin.PinTypes.PERIOD_PU _
                                                  OrElse p.GetPinType = Pin.PinTypes.USOUND_SENSOR Then
                Return True
            End If
        Next
        Return False
    End Function
End Class


' ==================================================================================================
'   SUBCLASS - Slave_MasterPinsV2
' ==================================================================================================
Class Slave_MasterPinsV2 : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.MasterPinsV2, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 10
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9))
    End Sub

    Friend Overrides Sub ConfigurePin(ByVal npin As Int32, ByVal pinType As Pin.PinTypes)
        Select Case pinType
            Case Pin.PinTypes.UNUSED
                Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.DIG_OUT, _
                 Pin.PinTypes.PWM_8, _
                 Pin.PinTypes.PWM_16, _
                 Pin.PinTypes.SERVO_8, _
                 Pin.PinTypes.SERVO_16, _
                 Pin.PinTypes.DIG_IN, _
                 Pin.PinTypes.DIG_IN_PU, _
                 Pin.PinTypes.COUNTER, _
                 Pin.PinTypes.COUNTER_PU, _
                 Pin.PinTypes.SLOW_PERIOD, _
                 Pin.PinTypes.SLOW_PERIOD_PU
                If ValidGenericPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.ADC_8, _
                 Pin.PinTypes.ADC_16, _
                 Pin.PinTypes.CAP_8, _
                 Pin.PinTypes.CAP_16, _
                 Pin.PinTypes.RES_8, _
                 Pin.PinTypes.RES_16
                If ValidAdcPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.STEPPER
                If ValidStepperPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.STEPPER_DIR
                If ValidStepperDirPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.PWM_FAST
                If ValidPwmFastPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.FAST_COUNTER, Pin.PinTypes.FAST_COUNTER_PU
                If ValidFastCounterPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.USOUND_SENSOR, Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                If ValidPeriodCounterPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)
        End Select
    End Sub

    Friend Overrides Sub FillTypeCombo(ByVal cmb As ComboBox, ByRef p As Pin)
        cmb.Items.Clear()
        '
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED))
        '
        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16))
        End If

        If ValidStepperPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER))
        End If
        If ValidStepperDirPin(p) Then
            cmb.Items.Clear()
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER_DIR))
        End If
        If ValidPwmFastPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_FAST))
        End If

        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU))
        End If

        If ValidAdcPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16))
        End If

        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU))
        End If

        If ValidFastCounterPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU))
        End If

        If ValidPeriodCounterPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR))
        End If

        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD_PU))
        End If
    End Sub

    Private Function ValidGenericPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId > 0 Then
            If Pins(pinSelected.PinId - 1).GetPinType() = Pin.PinTypes.STEPPER Then
                Return False
            End If
        End If
        Return True
    End Function

    Private Function ValidAdcPin(ByVal pinSelected As Pin) As Boolean
        If Not ValidGenericPin(pinSelected) Then Return False
        If pinSelected.PinId > 5 Then Return False
        Return True
    End Function

    Private Function ValidStepperPin(ByVal pinSelected As Pin) As Boolean
        If CountUsedPwmPins(pinSelected) > 4 Then Return False
        Select Case pinSelected.PinId
            Case 0, 2, 4, 6, 8 : Return True
        End Select
        Return False
    End Function

    Private Function ValidStepperDirPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId < 1 Then Return False
        If Pins(pinSelected.PinId - 1).GetPinType() <> Pin.PinTypes.STEPPER Then Return False
        Return True
    End Function

    Private Function ValidPwmFastPin(ByVal pinSelected As Pin) As Boolean
        If Not ValidGenericPin(pinSelected) Then Return False
        If CountUsedPwmPins(pinSelected) > 4 Then Return False
        Return True
    End Function

    Private Function ValidFastCounterPin(ByVal pinSelected As Pin) As Boolean
        If Not ValidGenericPin(pinSelected) Then Return False
        For Each p As Pin In Pins
            If p Is pinSelected Then Continue For
            If p.GetPinType = Pin.PinTypes.FAST_COUNTER OrElse p.GetPinType = Pin.PinTypes.FAST_COUNTER_PU Then
                Return False
            End If
        Next
        Return True
    End Function

    Private Function ValidPeriodCounterPin(ByVal pinSelected As Pin) As Boolean
        If Not ValidGenericPin(pinSelected) Then Return False
        For Each p As Pin In Pins
            If p Is pinSelected Then Continue For
            If p.GetPinType = Pin.PinTypes.PERIOD OrElse p.GetPinType = Pin.PinTypes.PERIOD_PU _
                                                  OrElse p.GetPinType = Pin.PinTypes.USOUND_SENSOR Then
                Return False
            End If
        Next
        Return True
    End Function

    Private Function CountUsedPwmPins(ByVal pinSelected As Pin) As Int32
        Dim n As Int32 = 0
        For Each p As Pin In Pins
            If Object.ReferenceEquals(p, pinSelected) Then Continue For
            If p.GetPinType() = Pin.PinTypes.STEPPER OrElse p.GetPinType() = Pin.PinTypes.PWM_FAST Then
                n += 1
            End If
        Next
        Return n
    End Function

End Class




' ==================================================================================================
'   SUBCLASS - Slave_MasterPinsV4
' ==================================================================================================
Class Slave_MasterPinsV4 : Inherits Slave

    Friend Sub New(ByVal _masterid As Int32, ByVal _slaveid As Int32)
        MyBase.new(SlaveTypes.MasterPinsV4, _masterid, _slaveid)
        InitDefaultPins()
        LoadCurrent_mA = 10
    End Sub

    Private Sub InitDefaultPins()
        Pins = New List(Of Pin)
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10))
        Pins.Add(New Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11))
    End Sub

    Friend Overrides Sub ConfigurePin(ByVal npin As Int32, ByVal pinType As Pin.PinTypes)
        Select Case pinType
            Case Pin.PinTypes.UNUSED
                Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.DIG_OUT, _
                 Pin.PinTypes.PWM_8, _
                 Pin.PinTypes.PWM_16, _
                 Pin.PinTypes.SERVO_8, _
                 Pin.PinTypes.SERVO_16, _
                 Pin.PinTypes.DIG_IN, _
                 Pin.PinTypes.DIG_IN_PU, _
                 Pin.PinTypes.COUNTER, _
                 Pin.PinTypes.COUNTER_PU, _
                 Pin.PinTypes.SLOW_PERIOD, _
                 Pin.PinTypes.SLOW_PERIOD_PU
                If ValidGenericPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.ADC_8, _
                 Pin.PinTypes.ADC_16, _
                 Pin.PinTypes.CAP_8, _
                 Pin.PinTypes.CAP_16, _
                 Pin.PinTypes.RES_8, _
                 Pin.PinTypes.RES_16
                If ValidAdcPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.STEPPER
                If ValidStepperPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.STEPPER_DIR
                If ValidStepperDirPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.PWM_FAST
                If ValidPwmFastPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.FAST_COUNTER, Pin.PinTypes.FAST_COUNTER_PU
                If ValidFastCounterPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.USOUND_SENSOR, Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                If ValidPeriodCounterPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.ENCODER_A, Pin.PinTypes.ENCODER_A_PU
                If ValidEncoderAPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)

            Case Pin.PinTypes.ENCODER_B, Pin.PinTypes.ENCODER_B_PU
                If ValidEncoderBPin(Pins(npin)) Then Pins(npin).SetPinType(pinType)
        End Select
    End Sub

    Friend Overrides Sub FillTypeCombo(ByVal cmb As ComboBox, ByRef p As Pin)
        cmb.Items.Clear()
        '
        cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED))
        '
        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16))
        End If

        If ValidStepperPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER))
        End If
        If ValidStepperDirPin(p) Then
            cmb.Items.Clear()
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER_DIR))
        End If
        If ValidPwmFastPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_FAST))
        End If

        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU))
        End If

        If ValidAdcPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16))
        End If

        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU))
        End If

        If ValidFastCounterPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU))
        End If

        If ValidPeriodCounterPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR))
        End If

        If ValidEncoderAPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_A))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_A_PU))
        End If
        If ValidEncoderBPin(p) Then
            cmb.Items.Clear()
            If Pins(p.PinId - 1).GetPinType = Pin.PinTypes.ENCODER_A Then
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_B))
            Else
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_B_PU))
            End If
        End If

        If ValidGenericPin(p) Then
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD))
            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD_PU))
        End If
    End Sub

    Private Function ValidGenericPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId > 0 Then
            If Pins(pinSelected.PinId - 1).GetPinType() = Pin.PinTypes.STEPPER Then
                Return False
            End If
            If Pins(pinSelected.PinId - 1).GetPinType() = Pin.PinTypes.ENCODER_A Then
                Return False
            End If
            If Pins(pinSelected.PinId - 1).GetPinType() = Pin.PinTypes.ENCODER_A_PU Then
                Return False
            End If
        End If
        Return True
    End Function

    Private Function ValidAdcPin(ByVal pinSelected As Pin) As Boolean
        If Not ValidGenericPin(pinSelected) Then Return False
        If pinSelected.PinId > 5 Then Return False
        Return True
    End Function

    Private Function ValidStepperPin(ByVal pinSelected As Pin) As Boolean
        If CountUsedPwmPins(pinSelected) > 4 Then Return False
        Select Case pinSelected.PinId
            Case 0, 2, 4, 6, 8 : Return True
        End Select
        Return False
    End Function

    Private Function ValidStepperDirPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId < 1 Then Return False
        If Pins(pinSelected.PinId - 1).GetPinType() <> Pin.PinTypes.STEPPER Then Return False
        Return True
    End Function

    Private Function ValidPwmFastPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId > 9 Then Return False
        If Not ValidGenericPin(pinSelected) Then Return False
        If CountUsedPwmPins(pinSelected) > 4 Then Return False
        Return True
    End Function

    Private Function ValidFastCounterPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId > 9 Then Return False
        If Not ValidGenericPin(pinSelected) Then Return False
        For Each p As Pin In Pins
            If p Is pinSelected Then Continue For
            If p.GetPinType = Pin.PinTypes.FAST_COUNTER OrElse p.GetPinType = Pin.PinTypes.FAST_COUNTER_PU Then
                Return False
            End If
        Next
        Return True
    End Function

    Private Function ValidPeriodCounterPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId > 9 Then Return False
        If Not ValidGenericPin(pinSelected) Then Return False
        For Each p As Pin In Pins
            If p Is pinSelected Then Continue For
            If p.GetPinType = Pin.PinTypes.PERIOD OrElse p.GetPinType = Pin.PinTypes.PERIOD_PU _
                                                  OrElse p.GetPinType = Pin.PinTypes.USOUND_SENSOR Then
                Return False
            End If
        Next
        Return True
    End Function

    Private Function ValidEncoderAPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId > 10 Then Return False
        Dim pt As Pin.PinTypes = Pins(pinSelected.PinId + 1).GetPinType()
        If pt = Pin.PinTypes.STEPPER Then Return False
        If pt = Pin.PinTypes.ENCODER_A Then Return False
        If pt = Pin.PinTypes.ENCODER_A_PU Then Return False
        If pinSelected.PinId > 0 Then
            pt = Pins(pinSelected.PinId - 1).GetPinType()
            If pt = Pin.PinTypes.STEPPER Then Return False
            If pt = Pin.PinTypes.ENCODER_A Then Return False
            If pt = Pin.PinTypes.ENCODER_A_PU Then Return False
        End If
        Return True
    End Function

    Private Function ValidEncoderBPin(ByVal pinSelected As Pin) As Boolean
        If pinSelected.PinId < 1 Then Return False
        Dim pt As Pin.PinTypes = Pins(pinSelected.PinId - 1).GetPinType()
        If pt <> Pin.PinTypes.ENCODER_A And _
           pt <> Pin.PinTypes.ENCODER_A_PU Then Return False
        Return True
    End Function

    Private Function CountUsedPwmPins(ByVal pinSelected As Pin) As Int32
        Dim n As Int32 = 0
        For Each p As Pin In Pins
            If Object.ReferenceEquals(p, pinSelected) Then Continue For
            If p.GetPinType() = Pin.PinTypes.STEPPER OrElse p.GetPinType() = Pin.PinTypes.PWM_FAST Then
                n += 1
            End If
        Next
        Return n
    End Function

End Class







